from __future__ import annotations

from collections.abc import Sequence

from tumfl.AST.BaseFunctionDefinition import BaseFunctionDefinition
from tumfl.AST.Expression.Name import Name
from tumfl.AST.Expression.Vararg import Vararg
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class LocalFunctionDefinition(Statement):
    """Definition for a local function"""

    def __init__(
        self,
        token: Token,
        function_name: Name,
        parameters: Sequence[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "LocalFunctionDefinition")
        self.function_name: Name = function_name
        self.parameters: list[Name | Vararg] = list(parameters)
        self.body: Block = body

    @staticmethod
    def from_base_definition(
        base: BaseFunctionDefinition, function_name: Name
    ) -> LocalFunctionDefinition:
        return LocalFunctionDefinition(
            base.token, function_name, base.parameters, base.body
        )



from logging import warning

from .AST import ASTNode, String
from .basic_walker import NoneWalker


class Config(NoneWalker):
    def __init__(self, replacements: dict[str, ASTNode], prefix: str = "$$"):
        super().__init__()
        self.replacements: dict[str, ASTNode] = replacements
        self.prefix: str = prefix

    def visit_String(self, node: String) -> None:
        if node.value.startswith(self.prefix):
            if node.value[len(self.prefix) :] in self.replacements:
                node.replace(self.replacements[node.value[len(self.prefix) :]])
            else:
                warning(f"Found placeholder with no assigned value: {node.value}")



from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Any, Optional

from tumfl.AST.ASTNode import ASTNode
from tumfl.AST.Expression.Expression import Expression
from tumfl.AST.Expression.Name import Name
from tumfl.Token import Token

from .Statement import Statement


class AttributedName:
    def __init__(self, name: Name, attribute: Optional[Name] = None):
        self.name: Name = name
        self.attribute: Optional[Name] = attribute

    def __repr__(self) -> str:
        return f"AttributedName(name={self.name!r}, attribute={self.attribute!r})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, AttributedName):
            return self.name == other.name and self.attribute == other.attribute
        return False

    def __str__(self) -> str:
        if self.attribute:
            return f"{self.name.variable_name} <{self.attribute.variable_name}>"
        return self.name.variable_name


class LocalAssign(Statement):
    """Assignment of local variables"""

    def __init__(
        self,
        token: Token,
        variable_names: Sequence[AttributedName],
        expressions: Optional[Sequence[Expression]],
    ):
        super().__init__(token, "LocalAssign")
        self.variable_names: list[AttributedName] = list(variable_names)
        self.expressions: Optional[list[Expression]] = (
            list(expressions) if expressions is not None else None
        )

    def parent(
        self, parent: Optional[ASTNode], file_name: Optional[Path] = None
    ) -> None:
        # due to attributed names not being AST nodes, the auto implementation does not work
        self.parent_class = parent
        for name in self.variable_names:
            name.name.parent(self)
            if name.attribute:
                name.attribute.parent(self)

        if self.expressions:
            for expr in self.expressions:
                expr.parent(self)



from __future__ import annotations

from collections.abc import Sequence

from tumfl.AST.BaseFunctionDefinition import BaseFunctionDefinition
from tumfl.AST.Statement.Block import Block
from tumfl.Token import Token

from .Expression import Expression
from .Name import Name
from .Vararg import Vararg


class ExpFunctionDefinition(Expression):
    """Definition of a lua function expression"""

    def __init__(
        self,
        token: Token,
        parameters: Sequence[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "ExpFunctionDefinition")
        self.parameters: list[Name | Vararg] = list(parameters)
        self.body: Block = body

    @staticmethod
    def from_base_definition(base: BaseFunctionDefinition) -> ExpFunctionDefinition:
        return ExpFunctionDefinition(base.token, base.parameters, base.body)
